import { Algebra, Comonad, Monad } from './interfaces.js'

/** @import { Apply, Functor, Morphism } from './types.js' */

/** @template T */
export class Identity extends Algebra(Monad, Comonad) {
  #value

  /**
   * @param {T} value 
   */
  constructor(value) {
    super()

    this.#value = value
  }

  /**
   * @template T
   * @param {T} value 
   */
  static of(value) {
    return id(value)
  }

  /**
   * @template U
   * @type {Functor<T>['map']}
   * @param {Morphism<T, U>} f 
   * @returns {Identity<U>}
   */
  map(f) {
    return id(f(this.#value))
  }

  /**
   * @template U
   * @type {Apply<T>['ap']}
   * @param {Apply<Morphism<T, U>>} b 
   * @returns {Apply<U>}
   */
  ap(b) {
    return /** @type {Apply<U>} */ (b.map(f => f(this.#value)))
  }

  /**
   * @template U
   * @param {Morphism<T, Identity<U>>} f 
   */
  chain(f) {
    return f(this.#value)
  }

  /**
   * @param {(value: Identity<T>) => T} f 
   */
  extend(f) {
    return id(f(this))
  }

  extract() {
    return this.#value
  }

  toString() {
    return `Identity(${this.#value})`
  }
}

/**
 * @template T
 * @param {T} value
 */
export const id = value => new Identity(value)

