import { id } from '../../vendor/izuna/src/index.js'
import { Algebra, Monad } from './interfaces.js'

/** @import { ApplicativeTypeRef, Apply, Chain, Functor, InferredMorphism, Morphism } from './types.js' */

/**
 * @template T
 * @typedef {InferredMorphism<T>} ReaderFn
 */

/** @template T */
export class Reader extends Algebra(Monad) {
  #run

  /** @param {InferredMorphism<T>} run */
  constructor(run) {
    super()

    this.#run = run
  }

  /**
   * @template T
   * @type {ApplicativeTypeRef<T, Reader<T>>['of']}
   * @param {T} a
   * @returns {Reader<T>}
   */
  static of(a) {
    return new Reader(/** @type {InferredMorphism<T>} */(_env => a))
  }

  /** @template T */
  static ask() {
    return new Reader(/** @type {InferredMorphism<T>} */(id))
  }

  /**
   * @type {Functor<T>['map']}
   * @param {InferredMorphism<T>} f 
   * @returns {Reader<T>}
   */
  map(f) {
    return this.chain(value => Reader.of(f(value)))
  }

  /**
   * @type {Chain<T>['chain']}
   * @param {InferredMorphism<T>} f 
   * @returns {Reader<T>}
   */
  chain(f) {
    return new Reader(env => {
      const result = this.#run(env)
      const next = f(result)
      return next.run(env)
    })
  }

  /**
   * @template U
   * @type {Apply<T>['ap']}
   * @param {Reader<Morphism<T, U>>} b
   * @returns {Reader<U>}
   */
  ap(b) {
    return /** @type {Reader<U>} */ (b.chain(f =>
      /** @type {Reader<U>} */(this.map(f))
    ))
  }

  /**
   * @param {InferredMorphism<T>} f 
   * @returns {Reader<T>}
   */
  local(f) {
    return new Reader(/** @type {InferredMorphism<T>} */(env => this.#run(f(env))))
  }

  /**
   * @template U
   * @param {T} env 
   * @returns {U}
   */
  run(env) {
    return this.#run(env)
  }
}

